/*
 * DDE.C
 *
 * Callback function the handles transactions from DDEML
 *
 * Copyright (c)1992 Microsoft Corporation, All Right Reserved
 *
 * Kraig Brockschmidt, Software Design Engineer
 * Microsoft Systems Developer Relations
 * One Microsoft Way
 * Redmond, WA  98052
 *
 * Internet  :  kraigb@microsoft.com
 * Compuserve:  70750,2344
 * Fax       :  (206)936-7329
 */

#define STRICT
#include <windows.h>
#include <ddeml.h>
#include "exchange.h"


PAPPVARS pAV;
BOOL     fTimeOut;



/*
 * SetDDECallbackPAV
 *
 * Purpose:
 *  Simple function to get the pAV (application variables) pointer
 *  into this module.  We have to use a global to store it here since
 *  there's no method for getting application-specific data into the
 *  DDE callback.
 *
 * Parameters:
 *  pAVSet          PAPPVARS from the application.
 *
 * Return Value:
 *  None
 */

void SetDDECallbackPAV(PAPPVARS pAVSet)
    {
    pAV=pAVSet;     //Set the global
    return;
    }





/*
 * DDETransactionHandler
 *
 * Purpose:
 *  Receives and processes transactions from DDEML calling other application
 *  code as necesary to perform the requested transaction.
 *
 * Parameters:
 *  xType           UINT transaction code.
 *  cfData          UINT clipboard format of any data received.
 *  hConv           HCONV handle of the conversation.  Most useful when
 *                  handling multiple conversations.
 *  hsz1, hsz2      HSZ string handles containing names and data.
 *  hData           HDDEDATA containing the data of format cfData.
 *  dwData1,dwData2 DWORD transaction-defined data.
 *
 * Return Value:
 *  HDDEDATA        Various values, either data handles, error codes,
 *                  or the like.
 */

HDDEDATA CALLBACK DDETransactionHandler(UINT xType, UINT cfData, HCONV hConv
                                        , HSZ hsz1, HSZ hsz2, HDDEDATA hData
                                        , DWORD dwData1, DWORD dwData2)
    {
    static BOOL fVisible=FALSE;
    static UINT cAdvise=0;
    BOOL        fTemp;
    int         i, cPeers;
    DWORD       cch;
    LPSTR       psz, pszT;

    switch (xType)
        {
        case XTYP_XACT_COMPLETE:
            ODS("DDE:  XTYP_XACT_COMPLETE Entry.");

            pAV->fXActPending=FALSE;

            //Errors are determined through hData.
            pAV->fAsyncError=(NULL!=hData);

            ODS("DDE:  XTYP_XACT_COMPLETE Confirmed.");
            break;


        case XTYP_CONNECT:      //Incoming
            /*
             * Since we registered our server names, we only have to check
             * the topic name here.  Registering a service name prevents
             * us from getting possibly bombarded with unwanted connection
             * attempts.
             *
             * We check the topic name by comparing the topic in hsz1 to
             * the topic we registered in pAV->hszTopic.  We also check
             * the flag in dwData2 to prevent connections to ourselves.
             */

            ODS("DDE:  XTYP_CONNECT Entry.");

            if ((BOOL)dwData2 || DdeCmpStringHandles(hsz1, pAV->hszTopic))
                break;

            ODS("DDE:  XTYP_CONNECT Confirmed.");
            return (HDDEDATA)(DWORD)TRUE;


        case XTYP_CONNECT_CONFIRM:
             ODSlu("DDE:  XTYP_CONNECT_CONFIRM Entry.  hConv=%lu", (DWORD)hConv);
             break;


        case XTYP_POKE:         //Incoming
            /*
             * We'll get a Poke in two situations.  First, a peer
             * has successfully called DdeConnect and wants to confirm
             * a true connection with us, not just the Network DDE subsystem.
             * In this case we simply add the connecting machine name
             * to the list of peers we're serving.
             *
             * If we're serving data to a peer, then they'll be sending
             * the second type of XTYP_POKE that contains their data.
             * We, as a server, don't make a connection to that client
             * to become a client ourselves.  We could, and receive data
             * through XTYP_ADVDATA, but that takes an extra conversation
             * handle.  So in this case, we get the equivalent of advise
             * data through a poke instead.
             */

            ODSlu("DDE:  XTYP_POKE Entry.  hConv=%lu", (DWORD)hConv);
            fTemp=FALSE;

            /*
             * Check if we're receiving a local name from a peer that
             * connected to us and is using us as a server.
             */
            if (!DdeCmpStringHandles(hsz2, pAV->hszItemLocal))
                {
                /*
                 * Items match, so get a pointer to it and the peer name
                 * store it with the conversation handle in pAV->hListS.
                 */
                psz=(LPSTR)DdeAccessData(hData, &cch);
                i=(int)SendMessage(pAV->hListS, LB_ADDSTRING, 0, (LONG)psz);
                SendMessage(pAV->hListS, LB_SETITEMDATA, i, (DWORD)hConv);
                DdeUnaccessData(hData);
                fTemp=TRUE;
                }

            /*
             * A poke from a client using us as a could be their changed
             * data, so we treat this exactly like we do XTYP_ADVDATA, only
             * we have to find the conversation handle in hListS instead of
             * hListC.  Note that we only do this check for two-way or
             * multi-way connections since they are the only scenarios where
             * well see this type of poke.
             */
            if (ONEWAY!=pAV->iDL && !DdeCmpStringHandles(hsz2, pAV->hszItemData))
                {
                /*
                 * Look through the list of client to us and find hConv.
                 * This step is really unnecessary for this application
                 * because all the peers exchange exactly the same data.
                 * If peers were sending structures that we had to maintain
                 * separately, then this step of finding which peer sent
                 * the data is a necessary one.
                 */
                cPeers=(int)SendMessage(pAV->hListS, LB_GETCOUNT, 0, 0L);

                for (i=0; i < cPeers; i++)
                    {
                    //Locate the conversation.
                    if (hConv==(HCONV)SendMessage(pAV->hListS, LB_GETITEMDATA, i, 0L))
                        {
                        /*
                         * Note, we don't really care what i or hConv is...
                         * BUT, if we kept separate structures, i would be
                         * our structure index.
                         */
                        DdeGetData(hData, (LPVOID)&pAV->xData, CBXDATA, 0L);

                        //Show the new data.
                        InvalidateRect(pAV->hWnd, NULL, TRUE);
                        UpdateWindow(pAV->hWnd);
                        fTemp=TRUE;
                        break;
                        }
                    }
                }


            if (!fTemp)
                return (HDDEDATA)DDE_FNOTPROCESSED;

            ODS("DDE:  XTYP_POKE Confirmed.");
            return (HDDEDATA)(DWORD)DDE_FACK;



        case XTYP_REQUEST:      //Outgoing
            /*
             * If we're currently processing a request list ourselves,
             * we don't have a complete list to give away, so make 'em wait.
             */
            if (pAV->fBlockRequest)
                return (HDDEDATA)DDE_FBUSY;

            //We only care about this on mutli-connects.
            if (MULTI!=pAV->iDL)
                return (HDDEDATA)DDE_FNOTPROCESSED;


            ODSlu("DDE:  XTYP_REQUEST Entry.  hConv=%lu", (DWORD)hConv);

            //Validate the item name for the peer list.
            if (DdeCmpStringHandles(hsz2, pAV->hszItemPeer))
                return (HDDEDATA)DDE_FNOTPROCESSED;

            /*
             * Go through BOTH our connections listbox and count the
             * length of the peer names, then copy those names into
             * a single long string delimited with backslashes.  The
             * +1 is to account for the delimeter.
             */

            cch=0L;
            cPeers=(int)SendMessage(pAV->hListC, LB_GETCOUNT, 0, 0L);

            for (i=0; i < cPeers; i++)
                cch += 1L+SendMessage(pAV->hListC, LB_GETTEXTLEN, i, 0L);

            cPeers=(int)SendMessage(pAV->hListS, LB_GETCOUNT, 0, 0L);

            for (i=0; i < cPeers; i++)
                cch += 1L+SendMessage(pAV->hListS, LB_GETTEXTLEN, i, 0L);


            hData=DdeCreateDataHandle(pAV->dwDDE, NULL, (DWORD)cch, 0L
                                      , pAV->hszItemPeer, CF_TEXT, 0);


            if (NULL==hData)
                return (HDDEDATA)DDE_FNOTPROCESSED;

            //If we have any items, copy them all.
            if (0L!=cch)
                {
                psz=(LPSTR)DdeAccessData(hData, &cch);

                //Copy connections we made.
                pszT=psz;
                cPeers=(int)SendMessage(pAV->hListC, LB_GETCOUNT, 0, 0L);

                for (i=0; i < cPeers; i++)
                    {
                    psz += SendMessage(pAV->hListC, LB_GETTEXT, i, (LONG)psz);
                    *psz++='\\';        //Delimeter
                    }

                //Copy connections made on us.
                pszT=psz;
                cPeers=(int)SendMessage(pAV->hListS, LB_GETCOUNT, 0, 0L);

                for (i=0; i < cPeers; i++)
                    {
                    psz += SendMessage(pAV->hListS, LB_GETTEXT, i, (LONG)psz);
                    *psz++='\\';        //Delimeter
                    }

                //Final null-termination, back up over that last backslash
                *(--psz)=0;

                ODSsz("XTYP_REQUEST String: %s<end>", pszT);
                }

            DdeUnaccessData(hData);

            ODS("DDE:  XTYP_REQUEST Confirmed.");
            return hData;


        case XTYP_ADVSTART:     //Incoming
            ODSlu("DDE:  XTYP_ADVSTART Entry.  hConv=%lu", (DWORD)hConv);

            //Check that the item requested is correct.
            if (DdeCmpStringHandles(hsz2, pAV->hszItemData))
                return (HDDEDATA)FALSE;

            if (0==cAdvise++)
                EnableServerMenuItems(pAV->hWnd, TRUE);

            ODS("DDE:  XTYP_ADVSTART Confirmed.");
            return (HDDEDATA)(DWORD)TRUE;


        case XTYP_ADVREQ:       //Outgoing
            ODSlu("DDE:  XTYP_ADVREQ Entry.  hConv=%lu", (DWORD)hConv);

            /*
             * Any request we get in this application will be for
             * our XPOSITION format on the ITEM_POSITION item, so we
             * don't bother validating the topic or item (that was
             * already done in XTYP_CONNECT and XTYP_ADVSTART.
             */

            hData=DdeCreateDataHandle(pAV->dwDDE, (LPVOID)&pAV->xData
                                      , CBXDATA, 0L, pAV->hszItemData
                                      , pAV->cfItem, 0);


            ODS("DDE:  XTYP_ADVREQ Confirmed.");
            return (HDDEDATA)hData;


        case XTYP_ADVDATA:      //Incoming
            ODSlu("DDE:  XTYP_ADVDATA Entry.  hConv=%lu", (DWORD)hConv);

            /*
             * We know the data we asked for, so we'll just find the
             * appropriate peer structure, copy the data, and update
             * the window.
             */

            /*
             * Look through our server list (hListC) and find hConv.
             * This step is really unnecessary for this application
             * because all the peers exchange exactly the same data.
             * If peers were sending structures that we had to maintain
             * separately, then this step of finding which peer sent
             * the data is a necessary one.
             */
            cPeers=(int)SendMessage(pAV->hListC, LB_GETCOUNT, 0, 0L);

            fTemp=FALSE;
            for (i=0; i < cPeers; i++)
                {
                //Locate the conversation.
                if (hConv==(HCONV)SendMessage(pAV->hListC, LB_GETITEMDATA, i, 0L))
                    {
                    /*
                     * Note, we don't really care what i or hConv is...
                     * BUT, if we kept separate structures, i would be
                     * our structure index.
                     */
                    DdeGetData(hData, (LPVOID)&pAV->xData, CBXDATA, 0L);

                    //Show the new data.
                    InvalidateRect(pAV->hWnd, NULL, TRUE);
                    UpdateWindow(pAV->hWnd);
                    break;
                    }
                }

            ODS("DDE:  XTYP_ADVDATA Confirmed.");
            return (HDDEDATA)(DWORD)DDE_FACK;



        case XTYP_ADVSTOP:      //Incoming
            ODSlu("DDE:  XTYP_ADVSTOP Entry.  hConv=%lu", (DWORD)hConv);

            if (0==--cAdvise)
                EnableServerMenuItems(pAV->hWnd, FALSE);

            ODS("DDE:  XTYP_ADVSTOP Confirmed.");
            break;


        case XTYP_DISCONNECT:   //Incoming
            ODSlu("DDE:  XTYP_DISCONNECT Entry.  hConv=%lu", (DWORD)hConv);

            //Find the peer that's disconnecting.
            cPeers=(int)SendMessage(pAV->hListS, LB_GETCOUNT, 0, 0L);

            for (i=0; i < cPeers; i++)
                {
                //Locate the conversation so we can delete the entry.
                if (hConv==(HCONV)SendMessage(pAV->hListS, LB_GETITEMDATA, i, 0L))
                    {
                   #ifdef DEBUG
                    cch=SendMessage(pAV->hListS, LB_GETTEXT, i, (LONG)(LPSTR)pAV->szPeer);
                    ODSsz("Disconnect notification from %s.", (LPSTR)pAV->szPeer);

                    if ((LPARAM)LB_ERR==cch)
                        {
                        ODSsz("Disconnect:  %s not in hListS for LB_GETTEXT.", (LPSTR)pAV->szPeer);
                        break;
                        }
                   #endif

                    SendMessage(pAV->hListS, LB_DELETESTRING, i, 0L);
                    }
                }

            ODS("DDE:  XTYP_DISCONNECT Confirmed.");
            break;

        default:
            break;
        }

    dwData1=0L;         //Prevent -W4 warnings for unreferenced parameters.
    cfData=0;

    return (HDDEDATA) NULL;
    }






/*
 * FDDETransactionWait
 *
 * Purpose:
 *  Enters a Peek/Translate/Dispatch message loop to process all messages
 *  to the application until XTYP_XACT_COMPLETE is received or we
 *  time out through the timer callback function.
 *
 * Parameters:
 *
 * Return Value:
 *  BOOL            TRUE if the transaction was successful, FALSE otherwise.
 */

BOOL FDDETransactionWait(PAPPVARS pAV)
    {
    MSG         msg;
    FARPROC     pfn;
    BOOL        fRet=FALSE;
    int         iTimer;

    pAV->fAsyncError=TRUE;
    pAV->fTimeout=FALSE;

    pfn=MakeProcInstance((FARPROC)TimeoutProc, pAV->hInst);
    iTimer=SetTimer(NULL, 0, 20000, (TIMERPROC)pfn);

    while (pAV->fXActPending)
        {
        if (pAV->fTimeout)
            break;

        if (PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
            /*
             * We need to disable any way for the application to close while
             * we're in this loop, so we trap WM_CLOSE and ignore that message.
             */
            if (WM_CLOSE==msg.message)
                {
                ODS("Ignoring WM_CLOSE received in FDDETransactionWait.\r\n");
                continue;
                }

            TranslateMessage(&msg);
            DispatchMessage(&msg);
            }
        else
            WaitMessage();
        }

    FreeProcInstance(pfn);
    KillTimer(NULL, iTimer);
    return pAV->fAsyncError;
    }



/*
 * TimeoutProc
 *
 * Purpose:
 *  Serves to notify the application that a timeout occurred while waiting
 *  for an asynchronous transaction to complete.
 *
 * Parameters:
 *  Standard.  We don't do anything with them.
 *
 * Return Value:
 *  None
 */

void FAR PASCAL TimeoutProc(HWND hWnd, UINT iMsg, UINT iTimer, DWORD dw)
    {
    ODS("TimeoutProc entered...timing out...");

    pAV->fTimeout=TRUE;
    pAV->fAsyncError=FALSE;
    return;
    }
